import abc
import enum

from gym.utils import EzPickle
import numpy as np

from magical.base_env import BaseEnv, ez_init
import magical.entities as en
import magical.geom as geom


class BaseClusterEnv(BaseEnv, abc.ABC):
    """There are blocks of many colours and types. You must arrange them into
    distinct clusters. Depending on the demo, cluster membership must either be
    determined by shape type or shape colour (but if it's determined by one
    characteristic in the demo then it should be independent of the other
    characteristic). There should be exactly one cluster for each value of the
    membership characteristic (e.g. if clustering on colour, there should be
    one blue cluster, one red cluster, etc.).

    This class should not be instantiated directly. Instead, you should use
    ClusterColourEnv or ClusterShapeEnv."""
    class ClusterBy(str, enum.Enum):
        """What characteristic should blocks be clustered by?"""
        COLOUR = 'colour'
        TYPE = 'type'
        # TODO: add a 'both' option! (will require another demo scenario)

    def __init__(
            self,
            # should we randomise assignment of colours to blocks, or use
            # default ordering?
            rand_shape_colour=False,
            # should we randomise assignment of types to blocks, or use default
            # ordering?
            rand_shape_type=False,
            # should we jitter the positions of blocks and the robot?
            rand_layout_minor=False,
            # should we fully randomise the positions of blocks and the robot?
            rand_layout_full=False,
            # should we randomise number of blocks? (this requires us to
            # randomise everything else, too)
            rand_shape_count=False,
            # which block characteristic do we want the user to pay attention
            # to for clustering? (colour vs. shape type)
            cluster_by=ClusterBy.COLOUR,
            **kwargs):
        super().__init__(**kwargs)
        self.rand_shape_colour = rand_shape_colour
        self.rand_shape_type = rand_shape_type
        self.rand_shape_count = rand_shape_count
        assert not (rand_layout_minor and rand_layout_full)
        self.rand_layout_minor = rand_layout_minor
        self.rand_layout_full = rand_layout_full
        self.cluster_by = cluster_by
        if self.rand_shape_count:
            assert self.rand_layout_full, \
                "if shape count is randomised then layout must also be " \
                "fully randomised"
            assert self.rand_shape_type, \
                "if shape count is randomised then shape type must also be " \
                "randomised"
            assert self.rand_shape_colour, \
                "if shape count is randomised then colour must be " \
                "randomised too"

    @classmethod
    def make_name(cls, suffix=None):
        pass

    def on_reset(self):
        # make the robot at default position (will be randomised at end if
        # rand_layout is true)
        robot = self._make_robot(*self.DEFAULT_ROBOT_POSE)

        # 3x blue & 2x of each other colour
        default_colours = self.DEFAULT_BLOCK_COLOURS
        # 3x pentagon & 2x of each other shape type
        default_shape_types = self.DEFAULT_BLOCK_SHAPES
        # these were generated by randomly scattering shapes about the chosen
        # default robot position and then rounding down values a bit
        default_poses = self.DEFAULT_BLOCK_POSES
        default_n_shapes = len(default_colours)

        if self.rand_shape_count:
            n_shapes = self.rng.randint(7, 10 + 1)
            # rand_shape_count=True implies rand_layout=True, so these MUST be
            # randomised at the end
            poses = [((0, 0), 0)] * n_shapes
        else:
            n_shapes = default_n_shapes
            # if rand_layout=True, these will be randomised at the end
            poses = default_poses

        if self.rand_shape_colour:
            # make sure we have at least one of each colour
            colours = list(en.SHAPE_COLOURS)
            colours.extend([
                self.rng.choice(en.SHAPE_COLOURS)
                for _ in range(n_shapes - len(colours))
            ])
            self.rng.shuffle(colours)
        else:
            colours = default_colours

        if self.rand_shape_type:
            # make sure we have at least one of each type, too
            shape_types = list(en.SHAPE_TYPES)
            shape_types.extend([
                self.rng.choice(en.SHAPE_TYPES)
                for _ in range(n_shapes - len(shape_types))
            ])
            self.rng.shuffle(shape_types)
        else:
            shape_types = default_shape_types

        assert len(poses) == n_shapes
        assert len(colours) == n_shapes
        assert len(shape_types) == n_shapes

        shape_ents = []
        for ((x, y), angle), colour, shape_type \
                in zip(poses, colours, shape_types):
            shape = self._make_shape(shape_type=shape_type,
                                     colour_name=colour,
                                     init_pos=(x, y),
                                     init_angle=angle)
            shape_ents.append(shape)
        self.add_entities(shape_ents)

        # make index mapping characteristic values to blocks
        if self.cluster_by == self.ClusterBy.COLOUR:
            c_values_list = np.asarray(colours, dtype='object')
            self.__characteristic_values = np.unique(c_values_list)
        elif self.cluster_by == self.ClusterBy.TYPE:
            c_values_list = np.asarray(shape_types, dtype='object')
            self.__characteristic_values = np.unique(c_values_list)
        else:
            raise NotImplementedError(
                f"don't know how to cluster by '{self.cluster_by}'")
        self.__blocks_by_characteristic = {}
        assert len(c_values_list) == len(shape_ents)
        for shape, c_value in zip(shape_ents, c_values_list):
            c_list = self.__blocks_by_characteristic.setdefault(c_value, [])
            c_list.append(shape)

        # as in match_regions.py, this should be added after all shapes so it
        # appears on top, but before layout randomisation so that it gets added
        # to the space correctly
        self.add_entities([robot])

        if self.rand_layout_full or self.rand_layout_minor:
            if self.rand_layout_full:
                pos_limit = rot_limit = None
            else:
                pos_limit = self.JITTER_POS_BOUND
                rot_limit = self.JITTER_ROT_BOUND
            geom.pm_randomise_all_poses(space=self._space,
                                        entities=[robot, *shape_ents],
                                        arena_lrbt=self.ARENA_BOUNDS_LRBT,
                                        rng=self.rng,
                                        rand_pos=True,
                                        rand_rot=True,
                                        rel_pos_linf_limits=pos_limit,
                                        rel_rot_limits=rot_limit)

        # set up index for lookups
        self.__ent_index = en.EntityIndex(shape_ents)

    def score_on_end_of_traj(self):
        # Compute centroids for each value of the relevant characteristic
        # (either colour or shape type). Also compute mean squared distance
        # from centroid for each block in the cluster.
        nvals = len(self.__characteristic_values)
        centroids = np.zeros((nvals, 2))
        for c_idx, c_value in enumerate(self.__characteristic_values):
            c_blocks = self.__blocks_by_characteristic.get(c_value)
            if not c_blocks:
                centroid = (0, 0)
            else:
                positions = np.asarray([(b.shape_body.position.x,
                                         b.shape_body.position.y)
                                        for b in c_blocks])
                centroid = np.mean(positions, axis=0)
            centroids[c_idx] = centroid

        # Now for each block compute whether squared distance to nearest
        # incorrect centroid. A block is correctly clustered if the true
        # centroid is closer than the next-nearest centroid by a margin of at
        # least min_margin * (mean variation within true centroid). This
        # rewards tight clusterings.
        min_margin = 2.0  # higher = more strict
        n_blocks = 0
        n_correct = 0
        for c_idx, c_value in enumerate(self.__characteristic_values):
            for block in self.__blocks_by_characteristic.get(c_value, []):
                n_blocks += 1
                block_pos = np.array([[
                    block.shape_body.position.x,
                    block.shape_body.position.y,
                ]])
                centroid_sses = np.sum((block_pos - centroids)**2, axis=1)
                indices = np.arange(len(self.__characteristic_values))
                true_sse, = centroid_sses[indices == c_idx]
                bad_sses = centroid_sses[indices != c_idx]
                nearest_bad_centroid = np.min(bad_sses)
                true_centroid_sse = centroid_sses[c_idx]
                margin = min_margin * true_centroid_sse
                n_correct += int(
                    np.sqrt(true_sse) < np.sqrt(nearest_bad_centroid) - margin)

        # rescale so that frac_correct <= thresh gives 0 score, frac_correct ==
        # 1.0 gives 1 score. I've found it's common to frac_correct ranging
        # from 0.2 up to 0.4 just from random init; this clipping process means
        # that random init gives close to 0 average score.
        frac_correct = float(n_correct) / max(n_blocks, 1)
        thresh = 0.75
        score = max(frac_correct - thresh, 0) / (1 - thresh)

        return score


class ClusterColourEnv(BaseClusterEnv, EzPickle):
    DEFAULT_ROBOT_POSE = ((0.71692, -0.34374), 0.83693)
    DEFAULT_BLOCK_COLOURS = [
        en.ShapeColour.BLUE,
        en.ShapeColour.BLUE,
        en.ShapeColour.BLUE,
        en.ShapeColour.GREEN,
        en.ShapeColour.GREEN,
        en.ShapeColour.RED,
        en.ShapeColour.YELLOW,
        en.ShapeColour.YELLOW,
    ]
    DEFAULT_BLOCK_SHAPES = [
        en.ShapeType.CIRCLE,
        en.ShapeType.STAR,
        en.ShapeType.SQUARE,
        en.ShapeType.PENTAGON,
        en.ShapeType.PENTAGON,
        en.ShapeType.SQUARE,
        en.ShapeType.STAR,
        en.ShapeType.PENTAGON,
    ]
    DEFAULT_BLOCK_POSES = [
        ((-0.5147, 0.14149), -0.38871),
        ((-0.1347, -0.71414), 1.0533),
        ((-0.74247, -0.097592), 1.1571),
        ((-0.077363, -0.42964), -0.64379),
        ((0.51978, 0.1853), -1.1762),
        ((-0.5278, -0.21642), 2.9356),
        ((-0.54039, 0.48292), 0.072818),
        ((-0.16761, 0.64303), -2.3255),
    ]

    @ez_init()
    def __init__(self, *args, **kwargs):
        super().__init__(*args,
                         cluster_by=BaseClusterEnv.ClusterBy.COLOUR,
                         **kwargs)

    @classmethod
    def make_name(cls, suffix=None):
        base_name = 'ClusterColour'
        return base_name + (suffix or '') + '-v0'


class ClusterShapeEnv(BaseClusterEnv, EzPickle):
    # demo variant
    DEFAULT_ROBOT_POSE = ((0.286, -0.202), -1.878)
    DEFAULT_BLOCK_COLOURS = [
        en.ShapeColour.YELLOW,
        en.ShapeColour.BLUE,
        en.ShapeColour.RED,
        en.ShapeColour.RED,
        en.ShapeColour.GREEN,
        en.ShapeColour.YELLOW,
        en.ShapeColour.BLUE,
        en.ShapeColour.GREEN,
    ]
    DEFAULT_BLOCK_SHAPES = [
        en.ShapeType.SQUARE,
        en.ShapeType.PENTAGON,
        en.ShapeType.PENTAGON,
        en.ShapeType.PENTAGON,
        en.ShapeType.CIRCLE,
        en.ShapeType.STAR,
        en.ShapeType.STAR,
        en.ShapeType.CIRCLE,
    ]
    DEFAULT_BLOCK_POSES = [
        ((-0.414, 0.297), -1.731),
        ((0.068, 0.705), 2.184),
        ((0.821, 0.220), 0.650),
        ((-0.461, -0.749), -2.673),
        ((0.867, -0.149), -2.215),
        ((-0.785, -0.140), -0.405),
        ((-0.305, -0.226), 1.341),
        ((0.758, -0.708), -2.140),
    ]

    @ez_init()
    def __init__(self, *args, **kwargs):
        super().__init__(*args,
                         cluster_by=BaseClusterEnv.ClusterBy.TYPE,
                         **kwargs)

    @classmethod
    def make_name(cls, suffix=None):
        base_name = 'ClusterShape'
        return base_name + (suffix or '') + '-v0'



import abc
import collections
import io
import warnings

import numpy as np
import pandas as pd
from statsmodels.stats.weightstats import DescrStatsW

from magical.benchmarks import DEMO_ENVS_TO_TEST_ENVS_MAP


class EvaluationProtocol(abc.ABC):
    """Common elements of the multi-task evaluation protocol. Users of this
    class provide a way to roll out K trajectories on a Gym environment with a
    given name. The class does aggregation of statistics in a sensible way."""
    _called_init = False

    def __init__(self, demo_env_name, n_rollouts):
        self.n_rollouts = n_rollouts
        self.demo_env_name = demo_env_name
        self.test_env_names = [
            demo_env_name,
            *DEMO_ENVS_TO_TEST_ENVS_MAP[demo_env_name],
        ]
        self._called_init = True

    @property
    @abc.abstractmethod
    def run_id(self):
        """Should produce a string identifier distinguishing the current
        model/algorithm/etc. This might be, e.g., a model snapshot path, or
        even just a name like "Multi-task GAIL". This identifier will be
        available in a `run_id` column of the DataFrame returned by
        `do_eval`."""
        pass

    @abc.abstractmethod
    def obtain_scores(self, env_name):
        """This method should perform `self.n_rollouts` rollouts on the
        environment named `env_name`, then return the `eval_score`s from each
        rollout. Should never return _fewer_ than `self.n_rollouts` scores; if
        it produces more scores, then the later ones will be ignored."""
        pass

    def do_eval(self, verbose=False):
        if not self._called_init:
            raise ValueError(
                "EvaluationProtocol.__init__() was not called. Did you "
                "include a super().__init__(…) call in your subclass?")

        stats_tuples = []

        for env_name in self.test_env_names:
            scores = self.obtain_scores(env_name)
            if len(scores) < self.n_rollouts:
                raise ValueError(f".obtain_scores() returned only "
                                 f"{len(scores)} scores, but we asked for "
                                 f"{self.n_rollouts} scores")
            if len(scores) > self.n_rollouts:
                # We want to be consistent. Extra scores will reduce width of
                # CI, but if some runs have more scores than others then user
                # might get confused and think that the configurations with
                # more runs have lower population variance (they may not).
                warnings.warn(
                    f"Asked for {self.n_rollouts} scores but got "
                    f"{len(scores)} scores instead. Will truncate to only "
                    f"consider the first {self.n_rollouts} scores.")
                scores = scores[:self.n_rollouts]
            mean = np.mean(scores)
            interval = DescrStatsW(scores).tconfint_mean(0.05, 'two-sided')
            std = np.std(scores, ddof=1)
            stats_tuples.append(
                (env_name, mean, interval[0], interval[1], std))

        records = [
            collections.OrderedDict([
                ('demo_env', self.demo_env_name),
                ('test_env', env_name),
                ('mean_score', mean_score),
                ('ci95_lower', ci95_lower),
                ('ci95_upper', ci95_upper),
                ('std_score', std),
                ('run_id', self.run_id),
            ]) for env_name, mean_score, ci95_lower, ci95_upper, std in
            stats_tuples
        ]
        frame = pd.DataFrame.from_records(records)

        if verbose:
            print(f"Final mean scores for '{self.run_id}':")
            relevant_stats = [
                'test_env',
                'mean_score',
                'ci95_lower',
                'ci95_upper',
            ]
            print(frame[relevant_stats])

        return frame


def latexify_results(eval_data, id_column='run_id'):
    """Take a data frame produced by `EvaluationProtocol.eval_data()` and
    produce a LaTeX table of results for this method. Will use the `run_id`
    column as an algorithm name (or to get algorithm names, if there's more
    than one algorithm present in the given data). You can override that by
    specifying the `id_column` keyword argument."""

    # Each column of the LaTeX table corresponds to a particular evaluation
    # environment, while each row corresponds to an algorithm. In contrast,
    # each row of the given Pandas frame is represents a series of rollouts by
    # one particular algorithm on one particular test configuration.

    test_envs = eval_data['test_env'].unique()
    col_names = [r'\textbf{%s}' % e for e in test_envs]
    alg_names = eval_data[id_column].unique()

    # write to buffer so we can use print()
    fp = io.StringIO()

    # prefix is just LaTeX table setup
    print(r"\centering", file=fp)
    print(r"\begin{tabular}{l@{\hspace{1em}}%s}" % ("c" * len(col_names)),
          file=fp)
    print(r"\toprule", file=fp)

    # first line: env names
    print(r'\textbf{Randomisation} & ', end='', file=fp)
    print(' & '.join(col_names), end='', file=fp)
    print('\\\\', file=fp)
    print(r'\midrule', file=fp)

    # next lines: actual results
    for alg_name in alg_names:
        alg_mask = eval_data[id_column] == alg_name
        stat_parts = []
        for env_name in test_envs:
            full_mask = alg_mask & (eval_data['test_env'] == env_name)
            relevant_rows = list(eval_data[full_mask].iterrows())
            if len(relevant_rows) != 1:
                raise ValueError(
                    f"got {len(relevant_rows)} rows corresponding to "
                    f"{id_column}={alg_name} and test_env={env_name}, but "
                    f"expected one (maybe IDs in column {id_column} aren't "
                    f"unique?)")
            (_, row), = relevant_rows
            std = row['std_score']
            stat_parts.append(f'{row["mean_score"]:.2f} ($\\pm$ {std:.2f})')
        print(r'\textbf{%s} & ' % alg_name, end='', file=fp)
        print(' & '.join(stat_parts), end='', file=fp)
        print('\\\\', file=fp)
        print(r'\bottomrule', file=fp)
        print(r'\end{tabular}', file=fp)

    return fp.getvalue()



from gym.utils import EzPickle

from magical.base_env import BaseEnv, ez_init
import magical.entities as en
import magical.geom as geom

DEFAULT_QUERY_COLOUR = en.ShapeColour.YELLOW
DEFAULT_QUERY_SHAPE = en.ShapeType.PENTAGON
DEFAULT_OUT_BLOCK_SHAPES = [
    en.ShapeType.PENTAGON,
    en.ShapeType.CIRCLE,
    en.ShapeType.CIRCLE,
    en.ShapeType.SQUARE,
    en.ShapeType.STAR,
    # last one is the target shape
    DEFAULT_QUERY_SHAPE,
]
DEFAULT_OUT_BLOCK_COLOURS = [
    en.ShapeColour.GREEN,
    en.ShapeColour.RED,
    en.ShapeColour.RED,
    en.ShapeColour.YELLOW,
    en.ShapeColour.BLUE,
    # last one is the target shape
    DEFAULT_QUERY_COLOUR,
]
DEFAULT_OUT_BLOCK_POSES = [
    ((-0.066751, 0.7552), -2.9266),
    ((-0.05195, 0.31468), 1.5418),
    ((0.57528, -0.46865), -2.2141),
    ((0.40594, -0.74977), 0.24582),
    ((0.45254, 0.3681), -1.0834),
    # last one is the target shape
    ((0.76849, -0.10652), 0.10028),
]
DEFAULT_ROBOT_POSE = ((-0.57, 0.25), 3.83)
DEFAULT_TARGET_REGION_XYHW = (-0.72, -0.22, 0.67, 0.72)
DEFAULT_QUERY_BLOCK_POSE = ((-0.33, -0.49), -0.51)


class FindDupeEnv(BaseEnv, EzPickle):
    """Task where robot has to find a block that is part of a group of
    duplicate blocks and bring it to the goal region. Every instance of this
    environment starts with one element of the duplicate group already inside a
    single goal region, and at least one element outside. Moving other
    duplicates into the goal region doesn't satisfy the goal.

    Details/corner cases: there's only ever one 'target' block in the goal
    region to start with. There will be at least one corresponding block with
    the *exact same attributes* outside of the goal region. There may be more
    than one corresponding block, and if that's the case then the robot only
    needs to fetch one of them (fetching more incurs no penalty)."""
    @ez_init()
    def __init__(self,
                 rand_colours=False,
                 rand_shapes=False,
                 rand_count=False,
                 rand_layout_minor=False,
                 rand_layout_full=False,
                 **kwargs):
        super().__init__(**kwargs)
        self.rand_colours = rand_colours
        self.rand_shapes = rand_shapes
        self.rand_count = rand_count
        self.rand_layout_minor = rand_layout_minor
        self.rand_layout_full = rand_layout_full
        if self.rand_count:
            assert self.rand_layout_full and self.rand_shapes \
                and self.rand_colours, "if shape count is randomised then " \
                "layout, shapes, and colours must be fully randomised too"

    @classmethod
    def make_name(cls, suffix=None):
        base_name = 'FindDupe'
        return base_name + (suffix or '') + '-v0'

    def on_reset(self):
        # make the robot (pose is randomised at end, if necessary)
        robot_pos, robot_angle = DEFAULT_ROBOT_POSE
        robot = self._make_robot(robot_pos, robot_angle)

        # sort out shapes and colours of blocks, including the target
        # shape/colour
        query_colour = DEFAULT_QUERY_COLOUR
        query_shape = DEFAULT_QUERY_SHAPE
        out_block_colours = DEFAULT_OUT_BLOCK_COLOURS
        out_block_shapes = DEFAULT_OUT_BLOCK_SHAPES
        n_out_blocks = len(DEFAULT_OUT_BLOCK_COLOURS)
        if self.rand_count:
            # have between 1 and 5 randomly generated blocks, plus one outside
            # block that always matches the query block
            n_out_blocks = self.rng.randint(1, 5 + 1) + 1
        n_distractors = n_out_blocks - 1

        if self.rand_colours:
            query_colour = self.rng.choice(en.SHAPE_COLOURS)
            out_block_colours = self.rng.choice(en.SHAPE_COLOURS,
                                                size=n_distractors).tolist()
            # last block always matches the query
            out_block_colours.append(query_colour)
        if self.rand_shapes:
            query_shape = self.rng.choice(en.SHAPE_TYPES)
            out_block_shapes = self.rng.choice(en.SHAPE_TYPES,
                                               size=n_distractors).tolist()
            out_block_shapes.append(query_shape)

        # create the target region
        target_region_xyhw = DEFAULT_TARGET_REGION_XYHW
        if self.rand_layout_minor or self.rand_layout_full:
            if self.rand_layout_minor:
                hw_bound = self.JITTER_TARGET_BOUND
            else:
                hw_bound = None
            target_hw = geom.randomise_hw(self.RAND_GOAL_MIN_SIZE,
                                          self.RAND_GOAL_MAX_SIZE,
                                          self.rng,
                                          current_hw=target_region_xyhw[2:],
                                          linf_bound=hw_bound)
            target_region_xyhw = (*target_region_xyhw[:2], *target_hw)
        sensor = en.GoalRegion(*target_region_xyhw, query_colour)
        self.add_entities([sensor])
        self.__sensor_ref = sensor

        # set up poses, as necessary
        query_block_pose = DEFAULT_QUERY_BLOCK_POSE
        out_block_poses = DEFAULT_OUT_BLOCK_POSES
        if self.rand_count:
            # this will be filled out when randomising the layout
            out_block_poses = [((0, 0), 0)] * n_out_blocks

        # The "outside blocks" are those which need to be outside the goal
        # region in the initial state. That includes the clone of the query
        # block that we intentionally inserted above.
        outside_blocks = []
        # we also keep a list of blocks that count as target blocks
        self.__target_set = set()
        for bshape, bcol, (bpos, bangle) in zip(out_block_shapes,
                                                out_block_colours,
                                                out_block_poses):
            new_block = self._make_shape(shape_type=bshape,
                                         colour_name=bcol,
                                         init_pos=bpos,
                                         init_angle=bangle)
            outside_blocks.append(new_block)
            if bcol == query_colour and bshape == query_shape:
                self.__target_set.add(new_block)
        self.add_entities(outside_blocks)
        # now add the query block
        assert len(query_block_pose) == 2
        query_block = self._make_shape(shape_type=query_shape,
                                       colour_name=query_colour,
                                       init_pos=query_block_pose[0],
                                       init_angle=query_block_pose[1])
        self.__target_set.add(query_block)
        self.add_entities([query_block])
        # these are shapes that shouldn't end up in the goal region
        self.__distractor_set = set(outside_blocks) - self.__target_set

        # add robot last, but before layout randomisation
        self.add_entities([robot])

        if self.rand_layout_minor or self.rand_layout_full:
            all_ents = (sensor, robot, *outside_blocks)
            if self.rand_layout_minor:
                # limit amount by which position and rotation can be randomised
                pos_limits = self.JITTER_POS_BOUND
                rot_limit = self.JITTER_ROT_BOUND
            else:
                # no limits, can randomise as much as needed
                assert self.rand_layout_full
                pos_limits = rot_limit = None
            # randomise rotations of all entities but goal region
            rand_rot = [False] + [True] * (len(all_ents) - 1)

            geom.pm_randomise_all_poses(self._space,
                                        all_ents,
                                        self.ARENA_BOUNDS_LRBT,
                                        rng=self.rng,
                                        rand_pos=True,
                                        rand_rot=rand_rot,
                                        rel_pos_linf_limits=pos_limits,
                                        rel_rot_limits=rot_limit,
                                        ignore_shapes=query_block.shapes)

            # randomise the query block last, since it must be mostly inside
            # the (randomly-placed) sensor region
            query_pos_limit = max(
                0,
                min(target_region_xyhw[2:]) / 2 - self.SHAPE_RAD / 2)
            if self.rand_layout_minor:
                query_pos_limit = min(self.JITTER_POS_BOUND, query_pos_limit)
            geom.pm_shift_bodies(self._space,
                                 query_block.bodies,
                                 position=sensor.bodies[0].position)
            geom.pm_randomise_pose(self._space,
                                   query_block.bodies,
                                   self.ARENA_BOUNDS_LRBT,
                                   ignore_shapes=sensor.shapes,
                                   rng=self.rng,
                                   rand_pos=True,
                                   rand_rot=True,
                                   rel_pos_linf_limit=query_pos_limit,
                                   rel_rot_limit=rot_limit)

        # block lookup index
        self.__block_index = en.EntityIndex([query_block, *outside_blocks])

    def score_on_end_of_traj(self):
        overlap_ents = self.__sensor_ref.get_overlapping_ents(
            com_overlap=True, ent_index=self.__block_index)
        n_overlap_targets = len(self.__target_set & overlap_ents)
        n_overlap_distractors = len(self.__distractor_set & overlap_ents)
        # what fraction of required targets are in the overlap set? (need 2;
        # more will not make difference)
        have_two_shapes = float(n_overlap_targets >= 2)
        if len(overlap_ents) == 0:
            contamination_rate = 0
        else:
            # what fraction of the overlap set are distractors?
            contamination_rate = n_overlap_distractors / len(overlap_ents)
        return have_two_shapes * (1 - contamination_rate)



import itertools as it

from gym.utils import EzPickle
import numpy as np

from magical import geom
from magical.base_env import BaseEnv, ez_init
import magical.entities as en

INLIER_RAD_MULT = 1.5
MAX_SEP_RADS = 3.5
MIN_BLOCKS = 3
MAX_BLOCKS = 4
DEFAULT_ROBOT_POSE = ((0.702, -0.255), 0.347)
DEFAULT_BLOCK_COLOURS = [
    en.ShapeColour.BLUE,
    en.ShapeColour.YELLOW,
    en.ShapeColour.RED,
    en.ShapeColour.GREEN,
]
DEFAULT_BLOCK_SHAPES = [
    en.ShapeType.STAR,
    en.ShapeType.CIRCLE,
    en.ShapeType.STAR,
    en.ShapeType.PENTAGON,
]
DEFAULT_BLOCK_POSES = [((0.790, -0.820), -0.721), ((-0.177, 0.383), -1.733),
                       ((-0.051, -0.128), 2.696), ((-0.292, -0.745), -0.159)]


def longest_line(points, inlier_dist, max_separation):
    """Identifies lines of blocks by performing RANSAC-style robust regression,
    but with two differences:

    1. It exhaustively searches over all pairs, instead of randomly sampling.
    2. It ensures that the distance between adjacent pairs of blocks along the
       returned line is less than max_separation.

    Returns number of points in the longest identified line."""
    points = np.asarray(points)
    assert points.ndim == 2 and points.shape[1] == 2
    npts = len(points)
    best = min(1, npts)
    for i in range(npts - 1):
        for j in range(i + 1, npts):
            # draw a line (p_i -> p_j) and find all the inliers for that line
            pi = points[i]
            offs = points - pi[None]
            pj_off = offs[j]
            pj_unit = pj_off / np.linalg.norm(pj_off)
            proj_lens = np.squeeze(offs @ pj_unit[:, None], axis=1)
            assert len(proj_lens) == len(points)
            dists = np.linalg.norm(offs - proj_lens[:, None] * pj_unit, axis=1)
            inlier_inds, = np.nonzero(dists <= inlier_dist)
            if len(inlier_inds) <= best:
                continue

            # now that we've found the inliers, find the largest subsequence of
            # inliers that are separated by a distance of at most
            # max_separation along the line
            inlier_proj_lens = proj_lens[inlier_inds]
            inlier_proj_lens.sort()
            seps = np.abs(np.diff(inlier_proj_lens))
            # find longest run of nearby shapes
            all_runs = it.groupby(seps <= max_separation)
            one_run_lens = [len(list(r)) for v, r in all_runs if v]
            max_run = max(one_run_lens, default=0) + 1
            if max_run > best:
                best = max_run

    return best


class MakeLineEnv(BaseEnv, EzPickle):
    """In this environment, the objective is to put all blocks in a straight
    line, arranged at any angle."""
    @ez_init()
    def __init__(self, rand_colours, rand_shapes, rand_count,
                 rand_layout_minor, rand_layout_full, **kwargs):
        super().__init__(**kwargs)
        self.rand_colours = rand_colours
        self.rand_shapes = rand_shapes
        self.rand_count = rand_count
        self.rand_layout_minor = rand_layout_minor
        self.rand_layout_full = rand_layout_full
        if self.rand_count:
            assert self.rand_layout_full and self.rand_shapes \
                and self.rand_colours, "if shape count is randomised then " \
                "layout, shapes, and colours must be fully randomised too"
        self.inlier_dist = self.SHAPE_RAD * INLIER_RAD_MULT
        self.max_sep = self.SHAPE_RAD * MAX_SEP_RADS

    @classmethod
    def make_name(cls, suffix=None):
        base_name = "MakeLine"
        return base_name + (suffix or '') + '-v0'

    def on_reset(self):
        robot_pos, robot_angle = DEFAULT_ROBOT_POSE
        robot = self._make_robot(robot_pos, robot_angle)

        block_shapes = DEFAULT_BLOCK_SHAPES
        block_colours = DEFAULT_BLOCK_COLOURS
        block_poses = DEFAULT_BLOCK_POSES
        if self.rand_count:
            n_blocks = self.rng.randint(MIN_BLOCKS, MAX_BLOCKS + 1)
            block_poses = block_poses[:1] * n_blocks
        else:
            n_blocks = len(block_shapes)
        if self.rand_colours:
            block_colours = self.rng.choice(en.SHAPE_COLOURS,
                                            size=n_blocks).tolist()
        if self.rand_shapes:
            block_shapes = self.rng.choice(en.SHAPE_TYPES,
                                           size=n_blocks).tolist()

        self._blocks = []
        for bshape, bcol, (bpos, bangle) in zip(block_shapes, block_colours,
                                                block_poses):
            new_block = self._make_shape(shape_type=bshape,
                                         colour_name=bcol,
                                         init_pos=bpos,
                                         init_angle=bangle)
            self._blocks.append(new_block)

        self.add_entities(self._blocks)
        self.add_entities([robot])

        if self.rand_layout_minor or self.rand_layout_full:
            all_ents = (robot, *self._blocks)
            if self.rand_layout_minor:
                pos_limits = self.JITTER_POS_BOUND
                rot_limit = self.JITTER_ROT_BOUND
            else:
                assert self.rand_layout_full
                pos_limits = rot_limit = None

            geom.pm_randomise_all_poses(self._space,
                                        all_ents,
                                        self.ARENA_BOUNDS_LRBT,
                                        rng=self.rng,
                                        rand_pos=True,
                                        rand_rot=True,
                                        rel_pos_linf_limits=pos_limits,
                                        rel_rot_limits=rot_limit)

    def score_on_end_of_traj(self):
        points = np.asarray([(b.shape_body.position.x, b.shape_body.position.y)
                             for b in self._blocks],
                            dtype='float64')
        line_len = longest_line(points, self.inlier_dist, self.max_sep)
        max_line_len = len(points)
        # 2 outliers = score of 0; 1 outlier = score of 0.5; 0 outliers = score
        # of 1
        min_line_len = max(max_line_len - 2, 2)
        score = max(line_len - min_line_len, 0) / (max_line_len - min_line_len)
        return score



#!/usr/bin/env python3
import os
from setuptools import find_packages, setup


def readme():
    """Load README for use as package's long description."""
    this_dir = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(this_dir, 'README.md'), 'r') as fp:
        return fp.read()


setup(
    name='magical-il',
    version='0.0.1alpha1',
    author='Sam Toyer',
    license='ISC',
    url='https://github.com/qxcv/magical/',
    description='MAGICAL is a benchmark suite for robust imitation learning',
    python_requires='>=3.7',
    long_description=readme(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        'pymunk~=5.6.0',
        'pyglet==1.5.*',
        'gym==0.17.*',
        'Click>=7.0',
        'numpy>=1.17.4',
        'cloudpickle>=1.2.2',
        'statsmodels>=0.10.2',
        'requests>=2.20.0,==2.*',
    ])



from gym.utils import EzPickle
import numpy as np

from magical import geom
from magical.base_env import BaseEnv, ez_init
import magical.entities as en

SMALL_POS_BOUND = 0.05
DEFAULT_ROBOT_POSE = ((0.058, 0.53), -2.13)
DEFAULT_GOAL_COLOUR = en.ShapeColour.BLUE
DEFAULT_GOAL_XYHW = (-0.62, -0.17, 0.76, 0.75)


class MoveToRegionEnv(BaseEnv, EzPickle):
    """Simple task where the robot merely has to move to a single coloured goal
    region."""
    @ez_init()
    def __init__(self,
                 rand_poses_minor=False,
                 rand_poses_full=False,
                 rand_goal_colour=False,
                 **kwargs):
        super().__init__(**kwargs)
        assert not (rand_poses_minor and rand_poses_full), \
            "cannot specify both 'rand_poses_minor' and 'rand_poses_full'"
        self.rand_poses_minor = rand_poses_minor
        self.rand_poses_full = rand_poses_full
        self.rand_goal_colour = rand_goal_colour

    @classmethod
    def make_name(cls, suffix=None):
        base_name = 'MoveToRegion'
        return base_name + (suffix or '') + '-v0'

    def on_reset(self):
        goal_xyhw = DEFAULT_GOAL_XYHW
        if self.rand_poses_minor or self.rand_poses_full:
            # randomise width and height of goal region
            # (unfortunately this has to be done before pose randomisation b/c
            # I don't have an easy way of changing size later)
            if self.rand_poses_minor:
                hw_bound = self.JITTER_TARGET_BOUND
            else:
                hw_bound = None
            sampled_hw = geom.randomise_hw(self.RAND_GOAL_MIN_SIZE,
                                           self.RAND_GOAL_MAX_SIZE,
                                           self.rng,
                                           current_hw=goal_xyhw[2:],
                                           linf_bound=hw_bound)
            goal_xyhw = (*goal_xyhw[:2], *sampled_hw)

        # colour the goal region
        if self.rand_goal_colour:
            goal_colour = self.rng.choice(
                np.asarray(en.SHAPE_COLOURS, dtype='object'))
        else:
            goal_colour = DEFAULT_GOAL_COLOUR

        # place the goal region
        assert len(goal_xyhw) == 4, goal_xyhw
        goal = en.GoalRegion(*goal_xyhw, goal_colour)
        self.add_entities([goal])
        self.__goal_ref = goal

        # now place the robot
        default_robot_pos, default_robot_angle = DEFAULT_ROBOT_POSE
        robot = self._make_robot(default_robot_pos, default_robot_angle)
        self.add_entities([robot])
        self.__robot_ent_index = en.EntityIndex([robot])

        if self.rand_poses_minor or self.rand_poses_full:
            if self.rand_poses_minor:
                # limit amount by which position and rotation can be randomised
                pos_limits = self.JITTER_POS_BOUND
                rot_limits = [None, self.JITTER_ROT_BOUND]
            else:
                # no limits, can randomise as much as needed
                assert self.rand_poses_full
                pos_limits = rot_limits = None

            geom.pm_randomise_all_poses(self._space,
                                        (self.__goal_ref, self._robot),
                                        self.ARENA_BOUNDS_LRBT,
                                        rng=self.rng,
                                        rand_pos=True,
                                        rand_rot=(False, True),
                                        rel_pos_linf_limits=pos_limits,
                                        rel_rot_limits=rot_limits)

    def score_on_end_of_traj(self):
        # this one just has a lazy binary reward
        dist, _ = self.__goal_ref.goal_shape.point_query(
            self._robot.robot_body.position)
        if dist <= 0:
            reward = 1.0
        else:
            reward = 0.0
        assert 0 <= reward <= 1
        return reward



"""Tools for defining and tracking 'physics variables' for the environment,
such as the mass of shapes, the 'motive force' of the robot, the power of its
grippers, and so on. The list of physics-related variables tracked here is not
exhaustive; variables have been added here primarily so that they can be
subject to dynamics randomisation in a consistent way. For instance, we want
all shapes to have the same mass in any given state, even if some dynamics
randomisation variants of a task might randomise that single number on
environment reset.

See `base_env.py` for actual values of the real physics variables used in the
MAGICAL environments.

(in hindsight this is wayyyyy more complicated than it will to be for the
foreseeable future, but I had fun making it :D)"""

import collections

# names that can't be used for
RESERVED_NAMES = set(['defaults', 'sample', 'variables'])
RESERVED_PREFIXES = ['_', 'rand_']


def _is_var_name(name):
    return name not in RESERVED_NAMES \
        and all(not name.startswith(p) for p in RESERVED_PREFIXES)


class _PhysicsVariablesMeta(type):
    """Metaclass for classes containing physics variables. When a new attribute
    of type `PhysVar` (and an appropriate name---see `_is_var_name()`) is added
    to a class, this metaclass records it in a `variables` dictionary, but
    removes it from the final attributes list. Note that when an actual
    instance of a physics variables class is constructed, it will still have
    attributes with the same names, but whose values are filled in with either
    defaults or randomly sampled values, as appropriate."""
    def __new__(cls, name, bases, classdict, _pbm_skip=False):
        if _pbm_skip:
            return super().__new__(cls, name, bases, classdict)

        new_classdict = dict()
        variables = collections.OrderedDict()
        for k, v in classdict.items():
            if not _is_var_name(k):
                new_classdict[k] = v
                continue
            if not isinstance(v, PhysVar):
                raise TypeError("variable '{k}' should be of type PhysVar")
            variables[k] = v
        rv = super().__new__(cls, name, bases, new_classdict)
        rv.variables = variables
        return rv


class PhysicsVariablesBase(metaclass=_PhysicsVariablesMeta, _pbm_skip=True):
    """Base class for physics variables with convenience methods for sampling
    random variables, instantiating from defaults, etc."""
    def __init__(self, *, _var_values):
        """Constructor for classes holding physics variable values. This should
        NOT be called directly. Instead, call the sample() or defaults()
        methods, as appropriate."""
        if _var_values.keys() != self.variables.keys():
            raise ValueError("must supply all & only given variable names")
        for k, v in _var_values.items():
            assert not isinstance(v, PhysVar), \
                f"are you sure '{k}' is meant to be a PhysVar?"
            setattr(self, k, v)

    @classmethod
    def defaults(cls):
        """Return a new instance of the class with default variable values."""
        return cls(
            _var_values={k: v.default
                         for k, v in cls.variables.items()})

    @classmethod
    def sample(cls, rng):
        """Return a new instance of the class with randomly sampled variable
        values."""
        return cls(
            _var_values={k: v.sample(rng)
                         for k, v in cls.variables.items()})

    def __repr__(self):
        pairs = ', '.join(f'{k}={getattr(self, k)}' for k in self.variables)
        return f'{self.__class__.__name__}({pairs})'


class PhysVar:
    """Container for the default value and randomisation bounds of a physical
    variable. Randomisation is always uniform."""
    def __init__(self, default, bounds):
        assert len(bounds) == 2, bounds
        lower, upper = bounds
        assert lower <= default <= upper, (lower, default, upper)
        self.default = default
        self.lower = lower
        self.upper = upper

    def sample(self, rng):
        result = rng.uniform(self.lower, self.upper)
        return result



"""Some kind of 'entity' abstraction for game world objects. The robot, the
different shapes in the environment, and the non-interacting coloured 'goal
regions' are all examples of entities."""

import abc
import enum
import math
import weakref

import numpy as np
import pymunk as pm
import pymunk.autogeometry as autogeom

import magical.geom as gtools
import magical.gym_render as r
from magical.style import (COLOURS_RGB, GOAL_LINE_THICKNESS,
                           ROBOT_LINE_THICKNESS, SHAPE_LINE_THICKNESS,
                           darken_rgb, lighten_rgb)

# #############################################################################
# Entity ABC
# #############################################################################


class Entity(abc.ABC):
    """Basic class for logical 'things' that can be displayed on screen and/or
    interact via physics."""
    @abc.abstractmethod
    def setup(self, viewer, space, phys_vars):
        """Set up entity graphics/physics usig a gym_render.Viewer and a
        pm.Space. Only gets called once."""
        self.shapes = []
        self.bodies = []
        self.viewer = weakref.proxy(viewer)
        self.space = weakref.proxy(space)
        self.phys_vars = weakref.proxy(phys_vars)

    def update(self, dt):
        """Do an logic/physics update at some (most likely fixed) time
        interval."""
    def pre_draw(self):
        """Do a graphics state update to, e.g., update state of internal
        `Geom`s. This doesn't have to be done at every physics time step."""
    def reconstruct_signature(self):
        """Produce signature necessary to reconstruct this entity in its
        current pose. This is useful for creating new scenarios out of existing
        world states.

        Returns: tuple of (cls, kwargs), where:
            cls (type): the class that should be used to construct the
                instance.
            kwargs (dict): keyword arguments that should be passed to the
                constructor for cls."""
        raise NotImplementedError(
            f"no .reconstruct_signature() implementation for object "
            f"'{self}' of type '{type(self)}'")

    def generate_group_id(self):
        """Generate a new, unique group ID. Intended to be called from
        `.setup()` when creating `ShapeFilter`s."""
        if not hasattr(self.space, '_group_ctr'):
            self.space._group_ctr = 999
        self.space._group_ctr += 1
        return self.space._group_ctr

    @staticmethod
    def format_reconstruct_signature(cls, kwargs):
        """String-format a reconstruction signature. Makes things as
        easy to cut-and-paste as possible."""
        prefix = "    "
        kwargs_sig_parts = []
        for k, v in sorted(kwargs.items()):
            v_str = str(v)
            if isinstance(v, pm.vec2d.Vec2d):
                v_str = "(%.5g, %.5g)" % (v.x, v.y)
            elif isinstance(v, float):
                v_str = "%.5g" % v
            part = f"{prefix}{k}={v_str}"
            kwargs_sig_parts.append(part)
        kwargs_sig = ",\n".join(kwargs_sig_parts)
        result = f"{cls.__name__}(\n{kwargs_sig})"
        return result

    def add_to_space(self, *objects):
        """For adding a body or shape to the Pymunk 'space'. Keeps track of
        shapes/bodies so they can be used later. Should be called instead of
        space.add()."""
        for obj in objects:
            self.space.add(obj)
            if isinstance(obj, pm.Body):
                self.bodies.append(obj)
            elif isinstance(obj, pm.Shape):
                self.shapes.append(obj)
            elif isinstance(obj, pm.Constraint):
                pass
            else:
                raise TypeError(
                    f"don't know how to handle object '{obj}' of type "
                    f"'{type(obj)}' in class '{type(self)}'")


# #############################################################################
# Helpers
# #############################################################################


class EntityIndex:
    def __init__(self, entities):
        """Build a reverse index mapping shapes to entities. Assumes that all
        the shapes which make up an entity are stored as attributes on the
        shape, or are attached to bodies which are stored as attributes on the
        shape.

        Args:
            entities ([Entity]): list of entities to process.

        Returns:
            ent_index (dict): dictionary mapping shapes to entities."""
        self._shape_to_ent = dict()
        self._ent_to_shapes = dict()
        for entity in entities:
            shapes = entity.shapes
            self._ent_to_shapes[entity] = shapes
            for shape in shapes:
                assert shape not in self._shape_to_ent, \
                    f"shape {shape} appears in {entity} and " \
                    f"{self._shape_to_ent[shape]}"
                self._shape_to_ent[shape] = entity

    def entity_for(self, shape):
        """Look up the entity associated with a particular shape. Raises
        `KeyError` if no known entity owns the shape."""
        return self._shape_to_ent[shape]

    def shapes_for(self, ent):
        """Return a set of shapes associated with the given entity. Raises
        `KeyError` if the given entity is not in the index."""
        return self._ent_to_shapes[ent]


# #############################################################################
# Robot entity
# #############################################################################


class RobotAction(enum.IntFlag):
    NONE = 0
    UP = 1
    DOWN = 2
    LEFT = 4
    RIGHT = 8
    OPEN = 16
    CLOSE = 32


# We need to have unique, consecutive identifiers for each action so that we
# can use standard tools for categorical action spaces. This list & the
# structures that follow allow us to map robot action flag combinations to
# unique integer identifiers, and vice versa.
ACTION_NUMS_FLAGS_NAMES = (
    # (unique act ID, (up/down, left/right, open/close), act name)
    (0,  (RobotAction.NONE, RobotAction.NONE,  RobotAction.OPEN),  'Open'),  # noqa: E501
    (1,  (RobotAction.UP,   RobotAction.NONE,  RobotAction.OPEN),  'UpOpen'),  # noqa: E501
    (2,  (RobotAction.DOWN, RobotAction.NONE,  RobotAction.OPEN),  'DownOpen'),  # noqa: E501
    (3,  (RobotAction.NONE, RobotAction.LEFT,  RobotAction.OPEN),  'LeftOpen'),  # noqa: E501
    (4,  (RobotAction.UP,   RobotAction.LEFT,  RobotAction.OPEN),  'UpLeftOpen'),  # noqa: E501
    (5,  (RobotAction.DOWN, RobotAction.LEFT,  RobotAction.OPEN),  'DownLeftOpen'),  # noqa: E501
    (6,  (RobotAction.NONE, RobotAction.RIGHT, RobotAction.OPEN),  'RightOpen'),  # noqa: E501
    (7,  (RobotAction.UP,   RobotAction.RIGHT, RobotAction.OPEN),  'UpRightOpen'),  # noqa: E501
    (8,  (RobotAction.DOWN, RobotAction.RIGHT, RobotAction.OPEN),  'DownRightOpen'),  # noqa: E501
    (9,  (RobotAction.NONE, RobotAction.NONE,  RobotAction.CLOSE), 'Close'),  # noqa: E501
    (10, (RobotAction.UP,   RobotAction.NONE,  RobotAction.CLOSE), 'UpClose'),  # noqa: E501
    (11, (RobotAction.DOWN, RobotAction.NONE,  RobotAction.CLOSE), 'DownClose'),  # noqa: E501
    (12, (RobotAction.NONE, RobotAction.LEFT,  RobotAction.CLOSE), 'LeftClose'),  # noqa: E501
    (13, (RobotAction.UP,   RobotAction.LEFT,  RobotAction.CLOSE), 'UpLeftClose'),  # noqa: E501
    (14, (RobotAction.DOWN, RobotAction.LEFT,  RobotAction.CLOSE), 'DownLeftClose'),  # noqa: E501
    (15, (RobotAction.NONE, RobotAction.RIGHT, RobotAction.CLOSE), 'RightClose'),  # noqa: E501
    (16, (RobotAction.UP,   RobotAction.RIGHT, RobotAction.CLOSE), 'UpRightClose'),  # noqa: E501
    (17, (RobotAction.DOWN, RobotAction.RIGHT, RobotAction.CLOSE), 'DownRightClose'),  # noqa: E501
)  # yapf: disable
ACTION_ID_TO_FLAGS = {
    act_id: flags
    for act_id, flags, _ in ACTION_NUMS_FLAGS_NAMES
}
FLAGS_TO_ACTION_ID = {
    flags: act_id
    for act_id, flags, _ in ACTION_NUMS_FLAGS_NAMES
}


def make_finger_vertices(upper_arm_len, forearm_len, thickness, side_sign):
    """Make annoying finger polygons coordinates. Corresponding composite shape
    will have origin at root of upper arm, with upper arm oriented straight
    upwards and forearm above it."""
    up_shift = upper_arm_len / 2
    upper_arm_vertices = gtools.rect_verts(thickness, upper_arm_len)
    forearm_vertices = gtools.rect_verts(thickness, forearm_len)
    # now rotate upper arm into place & then move it to correct position
    upper_start = pm.vec2d.Vec2d(side_sign * thickness / 2, upper_arm_len / 2)
    forearm_offset_unrot = pm.vec2d.Vec2d(-side_sign * thickness / 2,
                                          forearm_len / 2)
    rot_angle = side_sign * math.pi / 8
    forearm_trans = upper_start + forearm_offset_unrot.rotated(rot_angle)
    forearm_trans.y += up_shift
    forearm_vertices_trans = [
        v.rotated(rot_angle) + forearm_trans for v in forearm_vertices
    ]
    for v in upper_arm_vertices:
        v.y += up_shift
    upper_arm_verts_final = [(v.x, v.y) for v in upper_arm_vertices]
    forearm_verts_final = [(v.x, v.y) for v in forearm_vertices_trans]
    return upper_arm_verts_final, forearm_verts_final


class Robot(Entity):
    """Robot body controlled by the agent."""
    def __init__(self, radius, init_pos, init_angle, mass=1.0):
        self.radius = radius
        self.init_pos = init_pos
        self.init_angle = init_angle
        self.mass = mass
        self.rel_turn_angle = 0.0
        self.target_speed = 0.0
        # max angle from vertical on inner side and outer
        self.finger_rot_limit_outer = math.pi / 8
        self.finger_rot_limit_inner = 0.0

    def reconstruct_signature(self):
        cls = type(self)
        kwargs = dict(radius=self.radius,
                      init_pos=self.robot_body.position,
                      init_angle=self.robot_body.angle,
                      mass=self.mass)
        return cls, kwargs

    def setup(self, *args, **kwargs):
        super().setup(*args, **kwargs)

        # physics setup, starting with main body
        # signature: moment_for_circle(mass, inner_rad, outer_rad, offset)
        inertia = pm.moment_for_circle(self.mass, 0, self.radius, (0, 0))
        self.robot_body = body = pm.Body(self.mass, inertia)
        body.position = self.init_pos
        body.angle = self.init_angle
        self.add_to_space(body)

        # For control. The rough joint setup was taken form tank.py in the
        # pymunk examples.
        self.control_body = control_body = pm.Body(body_type=pm.Body.KINEMATIC)
        control_body.position = self.init_pos
        control_body.angle = self.init_angle
        self.add_to_space(control_body)
        pos_control_joint = pm.PivotJoint(control_body, body, (0, 0), (0, 0))
        pos_control_joint.max_bias = 0
        pos_control_joint.max_force = self.phys_vars.robot_pos_joint_max_force
        self.add_to_space(pos_control_joint)
        rot_control_joint = pm.GearJoint(control_body, body, 0.0, 1.0)
        rot_control_joint.error_bias = 0.0
        rot_control_joint.max_bias = 2.5
        rot_control_joint.max_force = self.phys_vars.robot_rot_joint_max_force
        self.add_to_space(rot_control_joint)

        # googly eye control bodies & joints
        self.pupil_bodies = []
        for eye_side in [-1, 1]:
            eye_mass = self.mass / 10
            eye_radius = self.radius
            eye_inertia = pm.moment_for_circle(eye_mass, 0, eye_radius, (0, 0))
            eye_body = pm.Body(eye_mass, eye_inertia)
            eye_body.angle = self.init_angle
            eye_joint = pm.DampedRotarySpring(body, eye_body, 0, 0.1, 3e-3)
            eye_joint.max_bias = 3.0
            eye_joint.max_force = 0.001
            self.pupil_bodies.append(eye_body)
            self.add_to_space(eye_body, eye_joint)

        # finger bodies/controls (annoying)
        finger_thickness = 0.25 * self.radius
        finger_upper_length = 1.1 * self.radius
        finger_lower_length = 0.7 * self.radius
        self.finger_bodies = []
        self.finger_motors = []
        finger_vertices = []
        finger_inner_vertices = []
        self.target_finger_angle = 0.0
        for finger_side in [-1, 1]:
            # basic finger body
            finger_verts = make_finger_vertices(
                upper_arm_len=finger_upper_length,
                forearm_len=finger_lower_length,
                thickness=finger_thickness,
                side_sign=finger_side)
            finger_vertices.append(finger_verts)
            finger_inner_verts = make_finger_vertices(
                upper_arm_len=finger_upper_length - ROBOT_LINE_THICKNESS * 2,
                forearm_len=finger_lower_length - ROBOT_LINE_THICKNESS * 2,
                thickness=finger_thickness - ROBOT_LINE_THICKNESS * 2,
                side_sign=finger_side)
            finger_inner_verts = [[(x, y + ROBOT_LINE_THICKNESS)
                                   for x, y in box]
                                  for box in finger_inner_verts]
            finger_inner_vertices.append(finger_inner_verts)
            # these are movement limits; they are useful below, but also
            # necessary to make initial positioning work
            if finger_side < 0:
                lower_rot_lim = -self.finger_rot_limit_inner
                upper_rot_lim = self.finger_rot_limit_outer
            if finger_side > 0:
                lower_rot_lim = -self.finger_rot_limit_outer
                upper_rot_lim = self.finger_rot_limit_inner
            finger_mass = self.mass / 8
            finger_inertia = pm.moment_for_poly(finger_mass,
                                                sum(finger_verts, []))
            finger_body = pm.Body(finger_mass, finger_inertia)
            if finger_side < 0:
                delta_finger_angle = upper_rot_lim
                finger_body.angle = self.init_angle + delta_finger_angle
            else:
                delta_finger_angle = lower_rot_lim
                finger_body.angle = self.init_angle + delta_finger_angle
            # position of finger relative to body
            finger_rel_pos = (finger_side * self.radius * 0.45,
                              self.radius * 0.1)
            finger_rel_pos_rot = gtools.rotate_vec(finger_rel_pos,
                                                   self.init_angle)
            finger_body.position = gtools.add_vecs(body.position,
                                                   finger_rel_pos_rot)
            self.add_to_space(finger_body)
            self.finger_bodies.append(finger_body)

            # pin joint to keep it in place (it will rotate around this point)
            finger_pin = pm.PinJoint(
                body,
                finger_body,
                finger_rel_pos,
                (0, 0),
            )
            finger_pin.error_bias = 0.0
            self.add_to_space(finger_pin)
            # rotary limit joint to stop it from getting too far out of line
            finger_limit = pm.RotaryLimitJoint(body, finger_body,
                                               lower_rot_lim, upper_rot_lim)
            finger_limit.error_bias = 0.0
            self.add_to_space(finger_limit)
            # motor to move the fingers around (very limited in power so as not
            # to conflict with rotary limit joint)
            finger_motor = pm.SimpleMotor(body, finger_body, 0.0)
            finger_motor.rate = 0.0
            finger_motor.max_bias = 0.0
            finger_motor.max_force = self.phys_vars.robot_finger_max_force
            self.add_to_space(finger_motor)
            self.finger_motors.append(finger_motor)

        # For collision. Main body circle. Signature: Circle(body, radius,
        # offset).
        robot_group = 1
        body_shape = pm.Circle(body, self.radius, (0, 0))
        body_shape.filter = pm.ShapeFilter(group=robot_group)
        body_shape.friction = 0.5
        self.add_to_space(body_shape)
        # the fingers
        finger_shapes = []
        for finger_body, finger_verts, finger_side in zip(
                self.finger_bodies, finger_vertices, [-1, 1]):
            finger_subshapes = []
            for finger_subverts in finger_verts:
                finger_subshape = pm.Poly(finger_body, finger_subverts)
                finger_subshape.filter = pm.ShapeFilter(group=robot_group)
                # grippy fingers
                finger_subshape.friction = 5.0
                finger_subshapes.append(finger_subshape)
            self.add_to_space(*finger_subshapes)
            finger_shapes.append(finger_subshapes)

        # graphics setup
        # draw a circular body
        circ_body_in = r.make_circle(radius=self.radius - ROBOT_LINE_THICKNESS,
                                     res=100)
        circ_body_out = r.make_circle(radius=self.radius, res=100)
        robot_colour = COLOURS_RGB['grey']
        dark_robot_colour = darken_rgb(robot_colour)
        light_robot_colour = lighten_rgb(robot_colour, 4)
        circ_body_in.set_color(*robot_colour)
        circ_body_out.set_color(*dark_robot_colour)

        # draw the two fingers
        self.finger_xforms = []
        finger_outer_geoms = []
        finger_inner_geoms = []
        for finger_outer_subshapes, finger_inner_verts, finger_side in zip(
                finger_shapes, finger_inner_vertices, [-1, 1]):
            finger_xform = r.Transform()
            self.finger_xforms.append(finger_xform)
            for finger_subshape in finger_outer_subshapes:
                vertices = [(v.x, v.y) for v in finger_subshape.get_vertices()]
                finger_outer_geom = r.make_polygon(vertices)
                finger_outer_geom.set_color(*robot_colour)
                finger_outer_geom.add_attr(finger_xform)
                finger_outer_geoms.append(finger_outer_geom)

            for vertices in finger_inner_verts:
                finger_inner_geom = r.make_polygon(vertices)
                finger_inner_geom.set_color(*light_robot_colour)
                finger_inner_geom.add_attr(finger_xform)
                finger_inner_geoms.append(finger_inner_geom)

        for geom in finger_outer_geoms:
            self.viewer.add_geom(geom)
        for geom in finger_inner_geoms:
            self.viewer.add_geom(geom)

        # draw some cute eyes
        eye_shapes = []
        self.pupil_transforms = []
        for x_sign in [-1, 1]:
            eye = r.make_circle(radius=0.2 * self.radius, res=20)
            eye.set_color(1.0, 1.0, 1.0)
            eye_base_transform = r.Transform().set_translation(
                x_sign * 0.4 * self.radius, 0.3 * self.radius)
            eye.add_attr(eye_base_transform)
            pupil = r.make_circle(radius=0.12 * self.radius, res=10)
            pupil.set_color(0.1, 0.1, 0.1)
            # The pupils point forward slightly
            pupil_transform = r.Transform()
            pupil.add_attr(r.Transform().set_translation(
                0, self.radius * 0.07))
            pupil.add_attr(pupil_transform)
            pupil.add_attr(eye_base_transform)
            self.pupil_transforms.append(pupil_transform)
            eye_shapes.extend([eye, pupil])
        # join them together
        self.robot_xform = r.Transform()
        robot_compound = r.Compound([circ_body_out, circ_body_in, *eye_shapes])
        robot_compound.add_attr(self.robot_xform)
        self.viewer.add_geom(robot_compound)

    def set_action(self, move_action):
        self.rel_turn_angle = 0.0
        self.target_speed = 0.0
        if move_action & RobotAction.UP:
            self.target_speed += 4.0 * self.radius
        if move_action & RobotAction.DOWN:
            self.target_speed -= 3.0 * self.radius
        if (move_action & RobotAction.UP) and (move_action & RobotAction.DOWN):
            self.target_speed = 0.0
        if move_action & RobotAction.LEFT:
            self.rel_turn_angle += 1.5
        if move_action & RobotAction.RIGHT:
            self.rel_turn_angle -= 1.5
        if (move_action & RobotAction.OPEN):
            # setting target for LEFT finger
            # (for right finger you'll need to flip across main axis)
            self.target_finger_angle = self.finger_rot_limit_outer
        elif move_action & RobotAction.CLOSE:
            self.target_finger_angle = -self.finger_rot_limit_inner

    def update(self, dt):
        # target heading
        self.control_body.angle = self.robot_body.angle + self.rel_turn_angle

        # target speed
        x_vel_vector = pm.vec2d.Vec2d(0.0, self.target_speed)
        vel_vector = self.robot_body.rotation_vector.cpvrotate(x_vel_vector)
        self.control_body.velocity = vel_vector

        # target finger positions, relative to body
        for finger_body, finger_motor, finger_side in zip(
                self.finger_bodies, self.finger_motors, [-1, 1]):
            rel_angle = finger_body.angle - self.robot_body.angle
            # for the left finger, the target angle is measured
            # counterclockwise; for the right, it's measured clockwise
            # (chipmunk is always counterclockwise)
            angle_error = rel_angle + finger_side * self.target_finger_angle
            target_rate = max(-1, min(1, angle_error * 10))
            if abs(target_rate) < 1e-4:
                target_rate = 0.0
            finger_motor.rate = target_rate

    def pre_draw(self):
        self.robot_xform.set_translation(*self.robot_body.position)
        self.robot_xform.set_rotation(self.robot_body.angle)
        for finger_xform, finger_body in zip(self.finger_xforms,
                                             self.finger_bodies):
            finger_xform.set_translation(*finger_body.position)
            finger_xform.set_rotation(finger_body.angle)
        for pupil_xform, pupil_body in zip(self.pupil_transforms,
                                           self.pupil_bodies):
            pupil_xform.set_rotation(pupil_body.angle - self.robot_body.angle)


class ArenaBoundaries(Entity):
    """Handles physics of arena boundaries to keep everything in one place."""
    def __init__(self, left, right, top, bottom, seg_rad=1):
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom
        self.seg_rad = seg_rad

    def setup(self, *args, **kwargs):
        super().setup(*args, **kwargs)

        # thick line segments around the edges
        arena_body = pm.Body(body_type=pm.Body.STATIC)
        rad = self.seg_rad
        points = [(self.left - rad, self.top + rad),
                  (self.right + rad, self.top + rad),
                  (self.right + rad, self.bottom - rad),
                  (self.left - rad, self.bottom - rad)]
        arena_segments = []
        for start_point, end_point in zip(points, points[1:] + points[:1]):
            segment = pm.Segment(arena_body, start_point, end_point, rad)
            segment.friction = 0.8
            arena_segments.append(segment)
        self.add_to_space(*arena_segments)

        width = self.right - self.left
        height = self.top - self.bottom
        arena_square = r.make_rect(width=width, height=height, filled=True)
        arena_bounds = r.make_rect(width=width, height=height, filled=False)
        # we don't really need this at the moment (2020-05-17) because the
        # middle of the arena is at (0,0) anyway; it will become useful if we
        # move the arena or something like that.
        # centre_xform = r.Transform(self.left + width / 2,
        #                            self.bottom + height / 2)
        # centre_xform.set_translation(self.left, self.top)
        # arena_square.add_attr(centre_xform)
        # arena_bounds.add_attr(centre_xform)
        # arena is white, bounds are dark grey
        arena_square.set_color(1, 1, 1)
        arena_bounds.set_color(*COLOURS_RGB['grey'])
        # also the bounds have same thickness as goals
        arena_bounds.add_attr(r.LineWidth(GOAL_LINE_THICKNESS))
        self.viewer.add_geom(arena_square)
        self.viewer.add_geom(arena_bounds)


# #############################################################################
# Pushable shapes
# #############################################################################


class ShapeType(str, enum.Enum):
    TRIANGLE = 'triangle'
    SQUARE = 'square'
    PENTAGON = 'pentagon'
    # hexagon is somewhat hard to distinguish from pentagon, and octagon is
    # very hard to distinguish from circle at low resolutions
    HEXAGON = 'hexagon'
    OCTAGON = 'octagon'
    CIRCLE = 'circle'
    STAR = 'star'


class ShapeColour(str, enum.Enum):
    RED = 'red'
    GREEN = 'green'
    BLUE = 'blue'
    YELLOW = 'yellow'


# limited set of types and colours to use for random generation
# (WARNING: not all benchmarks use the two arrays below! Some have used their
# own arrays so that changes to the base SHAPE_TYPES array don't break the
# benchmark's default shape layout.)
SHAPE_TYPES = np.asarray([
    ShapeType.SQUARE,
    ShapeType.PENTAGON,
    ShapeType.STAR,
    ShapeType.CIRCLE,
],
                         dtype='object')
SHAPE_COLOURS = np.asarray([
    ShapeColour.RED,
    ShapeColour.GREEN,
    ShapeColour.BLUE,
    ShapeColour.YELLOW,
],
                           dtype='object')


class Shape(Entity):
    """A shape that can be pushed around."""
    def __init__(self,
                 shape_type,
                 colour_name,
                 shape_size,
                 init_pos,
                 init_angle,
                 mass=0.5):
        self.shape_type = shape_type
        # this "size" can be interpreted in different ways depending on the
        # shape type, but area of shape should increase quadratically in this
        # number regardless of shape type
        self.shape_size = shape_size
        self.colour_name = colour_name
        self.colour = COLOURS_RGB[self.colour_name]
        self.init_pos = init_pos
        self.init_angle = init_angle
        self.mass = mass

    def reconstruct_signature(self):
        cls = type(self)
        kwargs = dict(shape_type=self.shape_type,
                      colour_name=self.colour_name,
                      shape_size=self.shape_size,
                      init_pos=self.shape_body.position,
                      init_angle=self.shape_body.angle,
                      mass=self.mass)
        return cls, kwargs

    def setup(self, *args, **kwargs):
        super().setup(*args, **kwargs)

        # Physics. This joint setup was taken form tank.py in the pymunk
        # examples.

        if self.shape_type == ShapeType.SQUARE:
            self.shape_body = body = pm.Body()
            body.position = self.init_pos
            body.angle = self.init_angle
            self.add_to_space(body)

            side_len = math.sqrt(math.pi) * self.shape_size
            shape = pm.Poly.create_box(
                body,
                (side_len, side_len),
                # slightly bevelled corners
                0.01 * side_len)
            # FIXME: why is this necessary? Do I need it for the others?
            shape.mass = self.mass
            shapes = [shape]
            del shape
        elif self.shape_type == ShapeType.CIRCLE:
            inertia = pm.moment_for_circle(self.mass, 0, self.shape_size,
                                           (0, 0))
            self.shape_body = body = pm.Body(self.mass, inertia)
            body.position = self.init_pos
            body.angle = self.init_angle
            self.add_to_space(body)
            shape = pm.Circle(body, self.shape_size, (0, 0))
            shapes = [shape]
            del shape
        elif self.shape_type == ShapeType.STAR:
            star_npoints = 5
            star_out_rad = 1.3 * self.shape_size
            star_in_rad = 0.5 * star_out_rad
            star_verts = gtools.compute_star_verts(star_npoints, star_out_rad,
                                                   star_in_rad)
            # create an exact convex decpomosition
            convex_parts = autogeom.convex_decomposition(
                star_verts + star_verts[:1], 0)
            star_hull = autogeom.to_convex_hull(star_verts, 1e-5)
            star_inertia = pm.moment_for_poly(self.mass, star_hull, (0, 0), 0)
            self.shape_body = body = pm.Body(self.mass, star_inertia)
            body.position = self.init_pos
            body.angle = self.init_angle
            self.add_to_space(body)
            shapes = []
            star_group = self.generate_group_id()
            for convex_part in convex_parts:
                shape = pm.Poly(body, convex_part)
                # avoid self-intersection with a shape filter
                shape.filter = pm.ShapeFilter(group=star_group)
                shapes.append(shape)
                del shape
        else:
            # these are free-form shapes b/c no helpers exist in Pymunk
            if self.shape_type == ShapeType.TRIANGLE:
                # shrink to make it look more sensible and easier to grasp
                factor = 0.8
                num_sides = 3
            elif self.shape_type == ShapeType.PENTAGON:
                factor = 1.0
                num_sides = 5
            elif self.shape_type == ShapeType.HEXAGON:
                factor = 1.0
                num_sides = 6
            elif self.shape_type == ShapeType.OCTAGON:
                factor = 1.0
                num_sides = 8
            else:
                raise NotImplementedError("haven't implemented",
                                          self.shape_type)
            side_len = factor * gtools.regular_poly_circ_rad_to_side_length(
                num_sides, self.shape_size)
            poly_verts = gtools.compute_regular_poly_verts(num_sides, side_len)
            inertia = pm.moment_for_poly(self.mass, poly_verts, (0, 0), 0)
            self.shape_body = body = pm.Body(self.mass, inertia)
            body.position = self.init_pos
            body.angle = self.init_angle
            self.add_to_space(body)
            shape = pm.Poly(body, poly_verts)
            shapes = [shape]
            del shape

        for shape in shapes:
            shape.friction = 0.5
            self.add_to_space(shape)

        trans_joint = pm.PivotJoint(self.space.static_body, body, (0, 0),
                                    (0, 0))
        trans_joint.max_bias = 0
        trans_joint.max_force = self.phys_vars.shape_trans_joint_max_force
        self.add_to_space(trans_joint)
        rot_joint = pm.GearJoint(self.space.static_body, body, 0.0, 1.0)
        rot_joint.max_bias = 0
        rot_joint.max_force = self.phys_vars.shape_rot_joint_max_force
        self.add_to_space(rot_joint)

        # Drawing
        if self.shape_type == ShapeType.SQUARE:
            geoms_inner = [r.make_square(side_len - 2 * SHAPE_LINE_THICKNESS)]
            geoms_outer = [r.make_square(side_len)]
        elif self.shape_type == ShapeType.CIRCLE:
            geoms_inner = [
                r.make_circle(radius=self.shape_size - SHAPE_LINE_THICKNESS,
                              res=100),
            ]
            geoms_outer = [r.make_circle(radius=self.shape_size, res=100)]
        elif self.shape_type == ShapeType.STAR:
            star_short_verts = gtools.compute_star_verts(
                star_npoints, star_out_rad - SHAPE_LINE_THICKNESS,
                star_in_rad - SHAPE_LINE_THICKNESS)
            short_convex_parts = autogeom.convex_decomposition(
                star_short_verts + star_short_verts[:1], 0)
            geoms_inner = []
            for part in short_convex_parts:
                geoms_inner.append(r.make_polygon(part))
            geoms_outer = []
            for part in convex_parts:
                geoms_outer.append(r.make_polygon(part))
        elif self.shape_type == ShapeType.OCTAGON \
                or self.shape_type == ShapeType.HEXAGON \
                or self.shape_type == ShapeType.PENTAGON \
                or self.shape_type == ShapeType.TRIANGLE:
            apothem = gtools.regular_poly_side_length_to_apothem(
                num_sides, side_len)
            short_side_len = gtools.regular_poly_apothem_to_side_legnth(
                num_sides, apothem - SHAPE_LINE_THICKNESS)
            short_verts = gtools.compute_regular_poly_verts(
                num_sides, short_side_len)
            geoms_inner = [r.make_polygon(short_verts)]
            geoms_outer = [r.make_polygon(poly_verts)]
        else:
            raise NotImplementedError("haven't implemented", self.shape_type)

        for g in geoms_inner:
            g.set_color(*self.colour)
        for g in geoms_outer:
            g.set_color(*darken_rgb(self.colour))
        self.shape_xform = r.Transform()
        shape_compound = r.Compound(geoms_outer + geoms_inner)
        shape_compound.add_attr(self.shape_xform)
        self.viewer.add_geom(shape_compound)

    def pre_draw(self):
        self.shape_xform.set_translation(*self.shape_body.position)
        self.shape_xform.set_rotation(self.shape_body.angle)


# #############################################################################
# Sensor region for pushing shapes into.
# #############################################################################


class GoalRegion(Entity):
    """A goal region that the robot should push certain shapes into. It's up to
    the caller to figure out exactly which shapes & call methods for collision
    checking/scoring."""
    def __init__(self, x, y, h, w, colour_name):
        self.x = x
        self.y = y
        assert h > 0, w > 0
        self.h = h
        self.w = w
        self.colour_name = colour_name
        self.base_colour = COLOURS_RGB[colour_name]

    def reconstruct_signature(self):
        kwargs = dict(x=self.goal_body.position[0] - self.w / 2,
                      y=self.goal_body.position[1] + self.h / 2,
                      h=self.h,
                      w=self.w,
                      colour_name=self.colour_name)
        return type(self), kwargs

    def setup(self, *args, **kwargs):
        super().setup(*args, **kwargs)

        # the space only needs a sensor body
        self.goal_body = pm.Body(body_type=pm.Body.STATIC)
        self.goal_shape = pm.Poly.create_box(self.goal_body, (self.w, self.h))
        self.goal_shape.sensor = True
        self.goal_body.position = (self.x + self.w / 2, self.y - self.h / 2)
        self.add_to_space(self.goal_body, self.goal_shape)

        # Making visual display: region should consist of very lightly shaded
        # rectangle, surrounded by darker stippled border. Ideally corners on
        # the rectangle should be rounded.
        self.rect_xform = r.Transform()
        self.rect_xform.set_translation(*self.goal_body.position)
        self.rect_xform.set_rotation(self.goal_body.angle)

        inner_colour = lighten_rgb(self.base_colour, times=2)
        inner_rect = r.make_rect(width=self.w, height=self.h, filled=True)
        inner_rect.set_color(*inner_colour)
        inner_rect.add_attr(self.rect_xform)
        self.viewer.add_geom(inner_rect)

        outer_colour = self.base_colour
        outer_rect = r.make_rect(width=self.w, height=self.h, filled=False)
        outer_rect.set_color(*outer_colour)
        outer_rect.add_attr(r.LineStyle(0x00FF))
        outer_rect.set_linewidth(250 * GOAL_LINE_THICKNESS)
        outer_rect.add_attr(self.rect_xform)
        self.viewer.add_geom(outer_rect)

    def get_overlapping_ents(self,
                             ent_index,
                             contained=False,
                             com_overlap=False):
        """Get all entities overlapping this region.

        Args:
            ent_index (EntityIndex): index of entities to query over.
            contained (bool): set this to True to only return entities that are
                fully contained in the regions. Otherwise, if this is False,
                all entities that overlap the region at all will be returned.

        Returns:
            ents ([Entity]): list of entities intersecting the current one."""

        # first look up all overlapping shapes
        shape_results = self.space.shape_query(self.goal_shape)
        overlap_shapes = {r.shape for r in shape_results}

        # if necessary, do total containment check on shapes
        if contained:
            # This does a containment check based *only* on axis-aligned
            # bounding boxes. This is valid if our goal region is an
            # axis-aligned bounding box, but could lead to false positives if
            # the goal region were a different shape, or if it was rotated.
            goal_bb = self.goal_shape.bb
            overlap_shapes = {
                s
                for s in overlap_shapes if goal_bb.contains(s.bb)
            }
        if com_overlap:
            goal_bb = self.goal_shape.bb
            overlap_shapes = {
                s
                for s in overlap_shapes
                if goal_bb.contains_vect(s.body.position)
            }

        # now look up all indexed entities that own at least one overlapping
        # shape
        relevant_ents = set()
        for shape in overlap_shapes:
            try:
                ent = ent_index.entity_for(shape)
            except KeyError:
                # shape not in index
                continue
            relevant_ents.add(ent)

        # if necessary, filter the entities so that only those with *all*
        # shapes within the region (or with COMs of all bodies in the region)
        # are included
        if contained or com_overlap:
            new_relevant_ents = set()
            for relevant_ent in relevant_ents:
                shapes = set(ent_index.shapes_for(relevant_ent))
                if shapes <= overlap_shapes:
                    new_relevant_ents.add(relevant_ent)
            relevant_ents = new_relevant_ents

        return relevant_ents

    def pre_draw(self):
        # just so we can see if the thing accidentally moves :-)
        self.rect_xform.set_translation(*self.goal_body.position)
        self.rect_xform.set_rotation(self.goal_body.angle)
